/***
 * Think ADL Parser
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: think@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.primitive.mp.skel;

import static org.objectweb.fractal.cecilia.adl.idl.remote.CSerializationHelper.writeDeserialization;
import static org.objectweb.fractal.cecilia.adl.idl.remote.CSerializationHelper.writeDeserializationDeclaration;
import static org.objectweb.fractal.cecilia.adl.idl.util.Util.getPrimitiveType;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.fractal.cecilia.adl.idl.ast.Field;
import org.objectweb.fractal.cecilia.adl.idl.ast.InterfaceDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.Method;
import org.objectweb.fractal.cecilia.adl.idl.ast.MethodContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.Parameter;
import org.objectweb.fractal.cecilia.adl.idl.ast.ParameterContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.PrimitiveType;
import org.objectweb.fractal.cecilia.adl.idl.ast.Type;
import org.objectweb.fractal.cecilia.adl.idl.ast.PrimitiveType.PrimitiveTypeEnum;
import org.objectweb.fractal.cecilia.adl.idl.remote.VarNameValidator;
import org.objectweb.fractal.cecilia.adl.idl.util.CUtil;
import org.objectweb.fractal.cecilia.adl.idl.util.Util;
import org.objectweb.fractal.cecilia.adl.interfaces.AbstractInterfaceAdapterImplementationVisitor;

/**
 * Generates the implementation of a RPC stub component.
 */
public class SkeletonImplementationVisitor
    extends
      AbstractInterfaceAdapterImplementationVisitor {

  @Override
  protected AbstractInterfaceAdapterImplmentationTask createImplementationTask(
      final InterfaceDefinition itf) {
    return new SkeletonImplementationTask(itf);
  }

  public static class SkeletonImplementationTask
      extends
        AbstractInterfaceAdapterImplmentationTask {

    private int              methodIdValue;
    private VarNameValidator localVars;
    private String           pushMsg;
    private String           data;
    private String           methodId;
    private String           result;

    /**
     * @param itf the {@link InterfaceDefinition} that is processed by this
     *            task.
     */
    public SkeletonImplementationTask(final InterfaceDefinition itf) {
      super(itf);
    }

    public void enterInterface(final InterfaceDefinition itf) throws Exception {
      localVars = new VarNameValidator();

      // Enter name of every method parameters in VarNameValidator to ensure
      // that local variables will not clash.
      for (final Method method : ((MethodContainer) itf).getMethods()) {
        for (final Parameter p : ((ParameterContainer) method).getParameters()) {
          localVars.validateName(p.getName());
        }
      }

      methodIdValue = 0;
      pushMsg = localVars.validateName("pushMsg");
      data = localVars.validateName("data");
      methodId = localVars.validateName("methodId");
      result = localVars.validateName("result");

      final String componentCName = typeNameProviderItf.getCTypeName();
      cw.appendln("#include \"minus/bare/generic/exportbind/api/serialize.h\"");

      cw.append("#define REQUIRED ((struct ").append(componentCName).append(
          "_t *)_this)->type.imported ").endl();
      cw.append("#define NEWSTYLE_COMPONENT").endl();

      cw.append("  struct ").append(componentCName).append("_instancedata {};")
          .endl();
      cw.appendln("#include \"cecilia.h\"");
      cw.appendln("#include \"stdio.h\"");
      cw.endl();
      cw.append("int ").append(componentCName).append("_push_push_method ")
          .append("(void *_this, Rminus_bare_generic_exportbind_api_Message *")
          .append(pushMsg).append(") {").endl();
      cw.append("char *").append(data).append(" = (char *)").append(pushMsg)
          .append("->data;").endl();
      cw.append("char ").append(methodId).append(" = *(").append(data).append(
          "++);").endl();
      cw.append("switch (").append(methodId).append(") {").endl();
    }

    public void leaveInterface(final InterfaceDefinition itf) throws Exception {
      cw.append("default : printf(\"Invalid methodId %d\", ").append(methodId)
          .append(");").endl();
      cw.appendln("}");
      cw.appendln("return minus_bare_generic_exportbind_api_Push_E_OK;");
      cw.appendln("}");

      cw.append("#undef NEWSTYLE_COMPONENT").endl();
    }

    public void visitField(final Field field) throws Exception {
      // nothing todo
    }

    public void visitMethod(final Method method) throws Exception {
      final Type returnType = Util.getContainedType((method));
      final Parameter[] parameters = ((ParameterContainer) method)
          .getParameters();
      final boolean isVoidMethod = (returnType instanceof PrimitiveType)
          && getPrimitiveType((PrimitiveType) returnType) == PrimitiveTypeEnum.VOID;
      final Map<Parameter, Object> paramContext = new HashMap<Parameter, Object>();
      final Object resultContext = null;

      cw.append("case ").append(methodIdValue++).append(" :").endl();
      cw.appendln("{");

      // declare parameter variables
      for (final Parameter p : parameters) {
        cw.append(CUtil.buildParameter(p)).append(";").endl();
      }

      // declaration related to parameter deserialization
      for (final Parameter p : parameters) {
        final Object ctx = writeDeserializationDeclaration(Util
            .getContainedType(p), p.getName(), cw, localVars);
        paramContext.put(p, ctx);
      }

      // deserialize parameters
      for (final Parameter p : parameters) {
        writeDeserialization(Util.getContainedType(p), p.getName(), data, cw,
            paramContext.get(p), true);
      }

      // invoke methods
      cw.append("CALL(REQUIRED.itf, ").append(method.getName());
      for (final Parameter p : parameters)
        cw.append(", ").append(p.getName());
      cw.append(");").endl();

      cw.appendln("break;");
      cw.appendln("}");
      cw.endl();
    }
  }
}
