/***
 * Think ADL Parser
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: think@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.primitive.mp.stub;

import static org.objectweb.fractal.cecilia.adl.idl.util.CUtil.buildMethodDefinition;
import static org.objectweb.fractal.cecilia.adl.idl.util.Util.getPrimitiveType;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.fractal.cecilia.adl.idl.ast.Field;
import org.objectweb.fractal.cecilia.adl.idl.ast.InterfaceDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.Method;
import org.objectweb.fractal.cecilia.adl.idl.ast.Parameter;
import org.objectweb.fractal.cecilia.adl.idl.ast.ParameterContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.PrimitiveType;
import org.objectweb.fractal.cecilia.adl.idl.ast.Type;
import org.objectweb.fractal.cecilia.adl.idl.ast.PrimitiveType.PrimitiveTypeEnum;
import org.objectweb.fractal.cecilia.adl.idl.remote.CSerializationHelper;
import org.objectweb.fractal.cecilia.adl.idl.remote.VarNameValidator;
import org.objectweb.fractal.cecilia.adl.idl.util.Util;
import org.objectweb.fractal.cecilia.adl.interfaces.AbstractInterfaceAdapterImplementationVisitor;

/**
 * Generates the implementation of a RPC stub component.
 */
public class StubImplementationVisitor
    extends
      AbstractInterfaceAdapterImplementationVisitor {

  @Override
  protected AbstractInterfaceAdapterImplmentationTask createImplementationTask(
      final InterfaceDefinition itf) {
    return new StubImplementationTask(itf);
  }

  public static class StubImplementationTask
      extends
        AbstractInterfaceAdapterImplmentationTask {

    private int methodId;

    /**
     * @param itf the {@link InterfaceDefinition} that is processed by this
     *            task.
     */
    public StubImplementationTask(final InterfaceDefinition itf) {
      super(itf);
    }

    public void enterInterface(final InterfaceDefinition itf) throws Exception {
      methodId = 0;
      final String componentCName = typeNameProviderItf.getCTypeName();
      cw.appendln("#include \"minus/bare/generic/exportbind/api/serialize.h\"");

      cw.append("#define REQUIRED ((struct ").append(componentCName).append(
          "_t *)_this)->type.imported ").endl();
      cw.append("#define NEWSTYLE_COMPONENT").endl();

      cw.append("  struct ").append(componentCName).append("_instancedata {};")
          .endl();
      cw.appendln("#include \"cecilia.h\"");
      cw.endl();
    }

    public void leaveInterface(final InterfaceDefinition itf) throws Exception {
      cw.append("#undef NEWSTYLE_COMPONENT").endl();
    }

    public void visitField(final Field field) throws Exception {
      // nothing todo
    }

    public void visitMethod(final Method method) throws Exception {
      final Type returnType = Util.getContainedType((method));
      final Parameter[] parameters = ((ParameterContainer) method)
          .getParameters();
      final boolean isVoidMethod = (returnType instanceof PrimitiveType)
          && getPrimitiveType((PrimitiveType) returnType) == PrimitiveTypeEnum.VOID;
      final Map<Parameter, Object> paramContext = new HashMap<Parameter, Object>();

      final VarNameValidator localVars = new VarNameValidator();
      for (final Parameter p : parameters)
        localVars.validateName(p.getName());

      final String pushMsg = localVars.validateName("pushMsg");
      final String data = localVars.validateName("data");
      final String msgSize = localVars.validateName("msgSize");

      final String methodName = new StringBuffer().append(
          typeNameProviderItf.getCTypeName()).append("_itf_").append(
          method.getName()).append("_method").toString();
      // append method prototype
      cw.append(buildMethodDefinition(method, false, methodName)).append(" {")
          .endl();

      // declare invokeMsg and replyMsg
      cw.append("Rminus_bare_generic_exportbind_api_Message *").append(pushMsg)
          .append(";").endl();
      cw.append("char *").append(data).append(";").endl();

      // declaration related to parameter serialization
      for (final Parameter p : parameters) {
        final Object ctx = CSerializationHelper.writeSerializationDeclaration(
            Util.getContainedType(p), p.getName(), cw, localVars);
        paramContext.put(p, ctx);
      }

      // compute pushMsg size
      cw.append("int ").append(msgSize).append(" = 1;").endl();
      for (final Parameter p : parameters) {
        CSerializationHelper.writeSizeOf(Util.getContainedType(p), p.getName(),
            msgSize, cw, paramContext.get(p));
      }

      // create pushMsg
      cw.append(pushMsg).append(" = ").append(
          "CALL(REQUIRED.message_manager, createMessage, ").append(msgSize)
          .append(");").endl();

      // initialize data pointer
      cw.append(data).append(" = (char *)pushMsg->data;").endl();

      // set method identifier
      cw.append("*(").append(data).append("++) = ").append(methodId++).append(
          ";").endl();

      // serialize parameters
      for (final Parameter p : parameters) {
        CSerializationHelper.writeSerialization(Util.getContainedType(p), p
            .getName(), data, cw, paramContext.get(p), true);
      }
      cw.endl();

      // send message
      cw.append("CALL(REQUIRED.push, push, ").append(pushMsg).append(");")
          .endl();

      cw.appendln("}");
      cw.endl();
    }
  }
}